/* 
    CMC RPC Client library
    Copyright (C) 2010 Dell, Inc

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/
#include <stdlib.h>
#include <stdarg.h>
#include <curl/curl.h>
#include <jansson.h>
#include "cmcrpc-c.h"

int
ssn_entry_list_rpc(int ssntype, const char *username, const char *ip, int *sidlist, int *numfound)
{
    int i;
    char *json;
    json_t *request = json_object();
    json_t *params = json_object();
    json_t *j_username;
    json_t *j_ip;
    json_t *j_resp;
    json_error_t error;
    CURLcode cret;
    cmcrpc_t *rpc;

    *numfound = 0;

    if (username) {
	j_username = json_string(username);
    }
    else j_username = json_null();


    if (ip) {
	j_ip = json_string(ip);
    }
    else j_ip = json_null();


    curl_global_init(CURL_GLOBAL_NOTHING);


    /* setup request */
    json_object_set_new(request, "method", json_string("ssn_entry_list"));
    json_object_set_new(request, "id", json_null());
    json_object_set_new(request, "params", params);

    /* setup request parameters */
    json_object_set_new(params, "ssntype", json_integer(ssntype));
    json_object_set_new(params, "username", j_username);
    json_object_set_new(params, "ipaddr", j_ip);
    
    json = json_dumps(request, 0);

    rpc = cmcrpc_new(0, json);
    if ((cret = cmcrpc_run(rpc)) != CURLE_OK) {
	goto done;
    }

    /* free request */
    /*free(json);
    json_decref(request);
    json_decref(params);
    */

    //fprintf(stderr, "%s", cmcrpc_data(rpc));

    /* Handle response */
    j_resp = json_loads(cmcrpc_data(rpc), &error);
    if (!j_resp) {
        /* TODO: handle invalid response */
        goto done;
    }

    /* pull out the json-rpc result and sidlist */
    json_t *j_result = json_object_get(j_resp, "result");
    json_t *j_sidlist = json_object_get(j_result, "sidlist");

    for (i=0; i<json_array_size(j_sidlist); i++) {
        sidlist[i] = json_integer_value(json_array_get(j_sidlist, i));
    }

    *numfound = json_integer_value(json_object_get(j_result, "numfound"));
    
    json_decref(j_result ) ;
    json_decref(j_sidlist);
    json_decref(j_resp);
done:



    free(json);
    json_decref(request);
    json_decref(params);
    json_decref(j_username);
    json_decref(j_ip);
    cmcrpc_free(rpc);

    curl_global_cleanup();

    return 0;
}

int 
ssn_close_rpc(unsigned int sid)
{
    json_t *request = json_object();
    json_t *params = json_object();
    char *json;
    CURLcode cret;
    cmcrpc_t *rpc;


    curl_global_init(CURL_GLOBAL_NOTHING);

    /* setup request */
    json_object_set_new(request, "method", json_string("ssn_close"));
    json_object_set_new(request, "id", json_null());
    json_object_set_new(request, "params", params);

    /* setup parameters */
    json_object_set_new(params, "sid", json_integer(sid));

    json = json_dumps(request, 0);

    rpc = cmcrpc_new(0, json);
    if ((cret = cmcrpc_run(rpc)) != CURLE_OK) {
	goto done;
    }

done:

    json_decref(params);
    json_decref(request);
    free(json);
    cmcrpc_free(rpc);

    curl_global_cleanup();

    return 0;

}

void 
ssn_log_rpc(int targetlog, const char *fmt, ...)
{
    va_list args;
    char *buffer;
    json_t *request = json_object();
    json_t *params = json_object();
    char *json;
    CURLcode cret;
    cmcrpc_t *rpc;

    /* create log message */
    va_start(args, fmt);
    if (vasprintf(&buffer, fmt, args) == -1) {
	return;
    }
    va_end(args);


    curl_global_init(CURL_GLOBAL_NOTHING);

    /* setup request */
    json_object_set_new(request, "method", json_string("ssn_log"));
    json_object_set_new(request, "id", json_null());
    json_object_set_new(request, "params", params);

    /* setup parameters */
    json_object_set_new(params, "msg", json_string(buffer));
    json_object_set_new(params, "targetlog", json_integer(targetlog));

    json = json_dumps(request, 0);

    rpc = cmcrpc_new(0, json);
    if ((cret = cmcrpc_run(rpc)) != CURLE_OK) {
	goto done;
    }

    /* ignore the reponse */

done:

    json_decref(params);
    json_decref(request);
    free(json);
    cmcrpc_free(rpc);
    if(buffer) free(buffer);
    curl_global_cleanup();
}

int
ssn_is_password_mode_rpc()
{
    json_t *request = json_object();
    json_t *j_resp;
    json_t *j_result;
    json_t *j_returncode;
    char *json;
    CURLcode cret;
    cmcrpc_t *rpc;
    int rc;
    json_error_t error;

    curl_global_init(CURL_GLOBAL_NOTHING);

    /* setup request */
    json_object_set_new(request, "method", json_string("ssn_is_password_mode"));
    json_object_set_new(request, "id", json_null());
    json_object_set_new(request, "params", json_null());

    json = json_dumps(request, 0);

    rpc = cmcrpc_new(0, json);
    if ((cret = cmcrpc_run(rpc)) != CURLE_OK) {
	goto done;
    }

    /* Handle response */
    j_resp = json_loads(cmcrpc_data(rpc),  &error);
    if (!j_resp) {
	fprintf(stderr, "%d\n%s\n", error.line, error.text);
        goto done;
    }

    /* pull out the json-rpc result */
    j_result = json_object_get(j_resp, "result");
    j_returncode = json_object_get(j_result, "returncode");
    rc = json_integer_value(j_returncode);
    json_decref(j_resp);
    json_decref(j_result );
    json_decref(j_returncode );

done:
  
    json_decref(request);
    free(json);
    cmcrpc_free(rpc);
    
    curl_global_cleanup();

    return rc;

}

int 
ssn_is_vendor_mode_rpc()
{
    json_t *request = json_object();
    json_t *j_resp;
    json_t *j_result;
    json_t *j_returncode;
    char *json;
    CURLcode cret;
    cmcrpc_t *rpc;
    int rc = 0;
    json_error_t error;


    curl_global_init(CURL_GLOBAL_NOTHING);

    /* setup request */
    json_object_set_new(request, "method", json_string("ssn_is_vendor_mode"));
    json_object_set_new(request, "id", json_null());
    json_object_set_new(request, "params", json_null());

    /* convert request object into a json formatted string */
    json = json_dumps(request, 0);

    rpc = cmcrpc_new(0, json);
    if ((cret = cmcrpc_run(rpc)) != CURLE_OK) {
	goto done;
    }

    /* Handle response */
    j_resp = json_loads(cmcrpc_data(rpc),  &error);
    if (!j_resp) {
	fprintf(stderr, "%d\n%s\n", error.line, error.text);
        goto done;
    }

    /* pull out the json-rpc result */
    j_result = json_object_get(j_resp, "result");
    j_returncode = json_object_get(j_result, "returncode");
    rc = json_integer_value(j_returncode);
    
    json_decref(j_result);
    json_decref(j_returncode);
    json_decref(j_resp);

done:

    json_decref(request);
    free(json);
    cmcrpc_free(rpc);

    curl_global_cleanup();

    return rc;
}
int ssn_check_dcw_rpc(int sid)
{
    json_t *request = json_object();
    json_t *j_resp;
    json_t *j_result;
    json_t *j_returncode;
    json_t *params = json_object();
    char *json;
    CURLcode cret;
    cmcrpc_t *rpc;
    int rc;
    json_error_t error;

    curl_global_init(CURL_GLOBAL_NOTHING);

    /* setup request */
    json_object_set_new(request, "method", json_string("ssn_check_dcw"));
    json_object_set_new(request, "id", json_null());
    json_object_set_new(request, "params", params);

    /* setup request parameters */
    json_object_set_new(params, "sid", json_integer(sid));


    /* convert request object into a json formatted string */
    json = json_dumps(request, 0);

    rpc = cmcrpc_new(0, json);
    if ((cret = cmcrpc_run(rpc)) != CURLE_OK) {
	goto done;
    }

    /* Handle response */
    j_resp = json_loads(cmcrpc_data(rpc), &error);
    if (!j_resp) {
	fprintf(stderr, "%d\n%s\n", error.line, error.text);
        goto done;
    }

    /* pull out the json-rpc result */
    j_result = json_object_get(j_resp, "result");
    j_returncode = json_object_get(j_result, "returncode");
    rc = json_integer_value(j_returncode);
    json_decref(j_resp);

done:
    json_decref(request);
    json_decref(params);
    free(json);
    cmcrpc_free(rpc);

    curl_global_cleanup();

    return rc;
}

