/*
 * Copyright (C) 2007 Dell Inc.
 * Jordan Hargrave <jordan_hargrave@dell.com>
 *
 * seepmtd.c: concatenate Serial EEPROM devices for MTD access
 *	    Allows mounting EEPROM as jffs2 filesystem
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/mtd/map.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/concat.h>
#include <linux/mtd/flashchip.h>
#include <linux/i2c.h>


static unsigned short normal_i2c[] = { 0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, I2C_CLIENT_END };

I2C_CLIENT_INSMOD_1(mtdseep);

struct seep_softc
{
	struct mtd_info     mtd;
	struct i2c_adapter *adapter;
	int		    flags;
	int		    page_size;
	int		    addr;
};
spinlock_t	    g_lock;

static int ee_size;
#ifdef __MODULE__
MODULE_PARM(ee_size, int, 65536);
#endif
MODULE_PARM_DESC(ee_size, "Set EEPROM size in bytes");

static int seep_mtd_read(struct mtd_info *mtd, loff_t from, size_t len, size_t *, u_char *buf);
static int seep_mtd_write(struct mtd_info *mtd, loff_t from, size_t len, size_t *, const u_char *buf);
static int seep_mtd_erase(struct mtd_info *mtd, struct erase_info *einf);
static void seep_mtd_sync(struct mtd_info *mtd);

static int seep_i2c_io(struct seep_softc *,int, void *, int, int);
static int seep_i2c_read(struct seep_softc *, int, size_t, size_t *, uint8_t *);
static int seep_i2c_write(struct seep_softc *, int, size_t, size_t *, uint8_t *);
static int seep_i2c_detect(struct i2c_adapter *, int, int kind);
static int seep_i2c_attach_adapter(struct i2c_adapter *);
static int seep_i2c_detach_client(struct i2c_client *);

/* Preserve first 256-bytes of chip for FRU data */
#define	FRU_SIZE		256
#define	MAX_I2C_RETRY		5
#define	MAX_WAIT_TIME		16
#define	SEEPROM_SIZE		(64*2*1024)	// 128K total (64K per SEEPROM)
#define	SEEPROM_PAGE_SIZE	128
#define	SEEPROM_ERASE_SIZE	(8*SEEPROM_PAGE_SIZE)

#if (FRU_SIZE > SEEPROM_ERASE_SIZE)  // total size = integral # of eraseblocks
	#error "FRU_SIZE exceeds 1 eraseblock size"
#endif
#define READ_CUR		0
#define READ_SET		1
#define WRITE_SET       	2
#define TWOBYTE_PART    	0x0080
#define MAX_EEPROM 		8

#undef DEBUG
#define DEBUG  0

#if DEBUG
#define dprintk printk

static void 
seep_util_dump(int addr, int len, uint8_t *buf)
{
	int rc, i,j, mj;

	for (i=0; i<len; i+=16) {
		mj = 16;
		if (mj > len-i)
			mj = len-i;
		
		printk("%.4x: ", addr+i);
		for (j=0; j<mj; j++) 
			printk("%.2x ", buf[i+j]);
		printk("  ");
		for (j=0; j<mj; j++) {
			rc = buf[i+j];
			if (rc < ' ' || rc > 'z') rc = '.';
			printk("%c", rc);
		}
		printk("\n");
	}
}
static void
seep_dump_i2c(struct i2c_msg *msg, int nmsg)
{
	int idx, jdx;

	for (idx=0; idx<nmsg; idx++) {
		printk("[%.2x:%c]", msg[idx].addr, msg[idx].flags ? 'R' : 'W');
		for (jdx=0; jdx<msg[idx].len; jdx++) {
			printk("%.2x ", msg[idx].buf[jdx]);
		}
	}
	printk("\n");
}

#else

#define dprintk(x...)
//#define seep_util_dump(x,y,z)
//#define seep_dump_i2c(x,y)
//#define __seep_i2c_write(sc, count, buf, addr) seep_i2c_io(sc, count, buf, WRITE_SET, addr)
#endif

/* Validate that writes were issued successfully */
static int
__seep_i2c_write(struct seep_softc *sc, int count, void *buf, int addr)
{
	int rc=-1;
	uint8_t rbuf[count];

	if (count > sc->page_size)
		printk("write too large!\n");
	memset(rbuf, 0, count);
	if (seep_i2c_io(sc, count, buf, WRITE_SET, addr) == 0) {
		if (seep_i2c_io(sc, count, rbuf, READ_SET, addr) == 0) {
			if (memcmp(buf, rbuf, count) == 0)
				rc = 0; // ok
			else
				printk("write fails for %x @ %x\n", count, addr);
		}
	}
	return rc;
}

////////////////////////////////////////////////////////////////

static struct i2c_driver seep_i2c = {

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,14)
	.driver = {
		.name = "seepmtd",
		.owner = THIS_MODULE,
	},
	.id = I2C_DRIVERID_EEPROM,
	.attach_adapter = seep_i2c_attach_adapter,
	.detach_client = seep_i2c_detach_client,
	.command = NULL,
#else
	.owner = THIS_MODULE,
	.name  = "seepmtd",
	.id    = I2C_DRIVERID_EEPROM,
	.flags = I2C_DF_NOTIFY,
	.attach_adapter = seep_i2c_attach_adapter,
	.detach_client  = seep_i2c_detach_client,
#endif
};

static int seep_num_mtd;
static struct mtd_info *seep_mtd[MAX_EEPROM+1];

/*==============================================================*
 * MTD functions
 *==============================================================*/

/* compute transfer length per aligned block: min=len, max=page_size+1 */
static int seep_align(long addr, size_t len, int page_size)
{
	int nc;

	nc = -addr & (page_size - 1);
	if (nc == 0)
		nc = page_size;
	if (nc > len)
		nc = len;

	return nc;
}

static void seep_mtd_sync(struct mtd_info *mtd)
{
	dprintk("seep_mtd_sync\n");
}

/* Erase MTD Block */
static int seep_mtd_erase(struct mtd_info *mtd, struct erase_info *erase)
{
	int rc = 0;
	size_t rlen = 0;

	struct seep_softc *sc = mtd->priv;
	uint8_t erv[erase->len];

	dprintk("seep_mtd_erase : %x %x\n", erase->addr, erase->len);
	memset(erv, 0xFF, erase->len);
	rc = seep_i2c_write(sc, erase->addr, erase->len, &rlen, erv);
	if (rc == 0)
		erase->state = MTD_ERASE_DONE;
	else
		erase->state = MTD_ERASE_FAILED;
	mtd_erase_callback(erase);

	return rc;
}

/* Read MTD Block */
static int seep_mtd_read(struct mtd_info *mtd, loff_t from, 
			size_t len, size_t *rlen, u_char *buf)
{
	int rc = 0;
	struct seep_softc *sc = mtd->priv;

	*rlen = 0;
	if (from >= sc->mtd.size) {
		*rlen = 0;
		printk("seep_mtd_read: %x:%llx %x %x\n", sc->addr, from, len, sc->mtd.size);
		return -EIO;
	}
	dprintk("seep_mtd_read: %x:%llx %x %x\n", sc->addr, from, len, sc->mtd.size);
	rc = seep_i2c_read(sc, from, len, rlen, buf);

	return rc;
}

/* Write MTD Block */
static int seep_mtd_write(struct mtd_info *mtd, loff_t from, 
			size_t len, size_t *rlen, const u_char *buf)
{
	int rc = 0;
	struct seep_softc *sc = mtd->priv;

	*rlen = 0;
	if (from >= sc->mtd.size) {
		printk("seep_mtd_write: %x:%llx %x %x\n", sc->addr, from, len, sc->mtd.size);
		*rlen = 0;
		return -EIO;
	}
	dprintk("seep_mtd_write: %x:%llx %x %x\n", sc->addr, from, len, sc->mtd.size);
	rc = seep_i2c_write(sc, from, len, rlen, (void *)buf);

	return rc;
}

/*==============================================================*
 * I2C functions
 *==============================================================*/
static int seep_i2c_transfer(struct i2c_adapter *adapter, struct i2c_msg *msg, int nmsg)
{
	int rc=0;
	uint8_t c;
	struct i2c_msg xmsg;
	int count;
	int wcount;

	xmsg.addr = msg[0].addr;
	xmsg.len = 1;
	xmsg.buf = &c;
	xmsg.flags = I2C_M_RD;

	for (count=0; count<MAX_I2C_RETRY; count++) {
		rc = i2c_transfer(adapter, msg, nmsg);
		if (msg[1].flags!=I2C_M_RD) mdelay(10);
		for (wcount=0; wcount<MAX_WAIT_TIME; wcount++) {
			if (i2c_transfer(adapter, &xmsg, 1) == 1) {
				break;
			}
			mdelay(10);
			printk("seep_i2c_transfer: waiting for completion %d\n", wcount);
		}
		if (rc == nmsg) {
			return 0;
		}
		printk("seep_i2c_transfer: no transfer %d, retry %d\n", rc,count);
	}
	printk("seep_i2c_transfer: error %d, too many retries!\n", rc);
	return -1;
}

/* Submit request to I2C bus
 *   mode = READ_SET/WRITE_SET : Set address and read/write
 *   mode = READ_CUR : Read from current address
 */
static int seep_i2c_io(struct seep_softc *sc, 
		       int len, void *buf, 
		       int mode, int ioaddr)
{
	int rc=0;
	struct i2c_msg msg[2];
	uint8_t mbuf[2+len];
	int nmsg;

	memset(mbuf, 0, sizeof(mbuf));
	memset(msg, 0, sizeof(msg));

	nmsg = 1;
	msg[0].buf = mbuf;
	msg[0].addr = sc->addr;
	if (mode == READ_CUR) {
		/* Read current position */
		msg[0].buf = buf;
		msg[0].len = len;
		msg[0].flags = I2C_M_RD;
	}
	else if (sc->flags & TWOBYTE_PART) {
		/* Handle 2-byte devices */
		mbuf[0] = ioaddr >> 8;
		mbuf[1] = ioaddr;
		msg[0].len = 2;
	}
	else {
		/* Handle 1-byte devices */
		mbuf[0] = ioaddr;
		msg[0].len = 1;
	}
	/* Set Write Buffer */
	if (mode == WRITE_SET) {
		memcpy(msg[0].buf+msg[0].len, buf, len);
		msg[0].len += len;
	}
	else if (mode == READ_SET) {
		msg[1].addr = sc->addr;
		msg[1].buf = buf;
		msg[1].len = len;
		msg[1].flags = I2C_M_RD;
		nmsg = 2;
	}

	if ((rc=seep_i2c_transfer(sc->adapter, msg, nmsg)) < 0) {
		printk("Error at address: %x,%x\n", msg[0].addr, ioaddr);
	}
	return rc;
}

static int 
seep_i2c_read(struct seep_softc *sc, int addr, size_t len, size_t *rlen, uint8_t *buf)
{
	int count;

	addr += FRU_SIZE;
	memset(buf, 0xCC, len);
	spin_lock(&g_lock);
	while (len > 0) {
		count = seep_align(addr, len, sc->page_size);

		dprintk("Read %x bytes at %x:%x\n", count, sc->addr, addr);
		if (seep_i2c_io(sc, count, buf, READ_SET, addr) < 0) {
			printk("seep_i2c_read: read error %x bytes, len %x at %x:%x\n", 
				count, len, sc->addr, addr);
			spin_unlock(&g_lock);
			return -EIO;
		}
		addr  += count;
		buf   += count;
		len   -= count;
		*rlen += count;
	}
	spin_unlock(&g_lock);
	return 0;
}

static int 
seep_i2c_write(struct seep_softc *sc, int addr, size_t len, size_t *rlen, uint8_t *buf)
{
	int count;

	addr += FRU_SIZE;
	spin_lock(&g_lock);
	while (len > 0) {
		count = seep_align(addr, len, sc->page_size);

		dprintk("Write %x bytes at %x:%x\n", count, sc->addr, addr);
		if (__seep_i2c_write(sc, count, buf, addr)) {
			printk("seep_i2c_write: write error %x bytes, len %x at %x:%x\n", 
				count, len, sc->addr, addr);
			spin_unlock(&g_lock);
			return -EIO;
		}
		addr  += count;
		buf   += count;
		len   -= count;
		*rlen += count;
	}
	spin_unlock(&g_lock);
	return 0;
}

/* Return size of EEPROM */
static int seep_i2c_size(struct seep_softc *sc)
{
	int size;
	uint8_t olddata[8];
	uint8_t tmpdata[1024];
	uint8_t testdata[8] = { 0xDE, 0xAD, 0xC0, 0xDE, 0xBE, 0xEF, 0xAD, 0xE0 };
	size_t rlen = 0;

	if (ee_size)
		return ee_size;

	/* Save old pattern, write new pattern */
	seep_i2c_read(sc, 0x00, 8, &rlen, olddata);

	memcpy(testdata, &sc, 4);
	seep_i2c_write(sc, 0x00, 8, &rlen, testdata);
	for (size=16; size<=SEEPROM_SIZE; size<<=1) {
		seep_i2c_read(sc, size, 8, &rlen, tmpdata);
		if (memcmp(tmpdata, testdata, 8) == 0)
			break;
	}
	dprintk("Calculated size=%x\n", size);
	seep_i2c_write(sc, 0x00, 8, &rlen, olddata);

	return size;
}

static int seep_i2c_detect(struct i2c_adapter *adapter, int address, int kind)
{
	struct seep_softc *sc;

	// Validate that this adapter supports I2C and 
	// is not on 1st IBM bus (CPU Timing EEPROM)

	dprintk("seep_i2c_detect: %s %d\n", adapter->name, adapter->nr);
	if (!i2c_check_functionality(adapter, I2C_FUNC_I2C)) {
		return -ENODEV;
	}
	if (strcmp(adapter->name, "IBM IIC") || adapter->nr == 0) {
		return -ENODEV;
	}
	dprintk("seep_detect: %d:%s:%x\n", adapter->nr, adapter->name, address);

	sc = kmalloc(sizeof(struct seep_softc), GFP_KERNEL);
	if (sc == NULL) {
		printk("seep_i2c_detect: *** %s %d\n", adapter->name, adapter->nr);
		return -ENOMEM;
	}

	/* Initialize EEPROM variables */
	memset(sc, 0, sizeof(struct seep_softc));
	sc->adapter = adapter;
	sc->addr = address;
	sc->page_size = SEEPROM_PAGE_SIZE;
	sc->flags = TWOBYTE_PART;

	sc->mtd.name = "seepmtd";
	sc->mtd.type = MTD_RAM;
	sc->mtd.flags = MTD_WRITEABLE; //MTD_CLEAR_BITS|MTD_SET_BITS|MTD_ERASEABLE;

	sc->mtd.erasesize = SEEPROM_ERASE_SIZE;
	sc->mtd.erase = seep_mtd_erase;
	sc->mtd.read = seep_mtd_read;
	sc->mtd.write = seep_mtd_write;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,14)
	sc->mtd.writesize = SEEPROM_PAGE_SIZE;
#endif
	sc->mtd.sync = seep_mtd_sync;
	sc->mtd.size = (SEEPROM_SIZE/2) - SEEPROM_ERASE_SIZE; // set to default size
	sc->mtd.priv = sc;

	seep_mtd[++seep_num_mtd] = &sc->mtd;
	printk("Detected %x byte EEPROM at i2c-%x-%x\n", 
	       sc->mtd.size,
	       adapter->nr,
	       address);
	return 0;
}

static int seep_i2c_attach_adapter(struct i2c_adapter *adapter)
{
	dprintk("seep_attach_adapter\n");
	return i2c_probe(adapter, &addr_data, seep_i2c_detect);
}

static int seep_i2c_detach_client(struct i2c_client *client)
{
	dprintk("seep_detach_client\n");
	return 0;
}

/*==============================================================*
 * Entry points
 *==============================================================*/

static int __init seep_init(void)
{
	dprintk("seep_init\n");
	spin_lock_init(&g_lock);
	i2c_add_driver(&seep_i2c);
	dprintk("seep_init: Detected %d mtd devices\n", seep_num_mtd);
	if (seep_num_mtd != 0) {
		seep_mtd[0] = mtd_concat_create(&seep_mtd[1], seep_num_mtd, "SEEPROM MTD");
		if (seep_mtd[0] != NULL) {
			add_mtd_device(seep_mtd[0]);
		}
	}
	return 0;
}

static void __exit seep_exit(void)
{
	int idx;

	dprintk("seep_exit\n");
	if (seep_mtd[0] != NULL) {
		del_mtd_device(seep_mtd[0]);
		mtd_concat_destroy(seep_mtd[0]);
	}
	for (idx=1; idx<=seep_num_mtd; idx++) 
		kfree(seep_mtd[idx]->priv);
	i2c_del_driver(&seep_i2c);
}

module_init(seep_init);
module_exit(seep_exit);

MODULE_AUTHOR("Jordan Hargrave <jordan_hargrave@dell.com>");
MODULE_DESCRIPTION("MTD Driver for SEEPROM");
MODULE_LICENSE("GPL");

